#!/usr/bin/env python3

import argparse
import colorama
import sys
import os

from . import check_file

success = True

colorama.init(autoreset=True)

def process(filename, options={}):
    global success
    print(colorama.Style.BRIGHT+f'--- Processing {filename}')
    if not check_file(filename, options):
        success = False
        print(colorama.Fore.RED+f'{filename} is potentially unsafe')

def main():
    global success
    progname = os.path.basename(sys.argv[0])
    parser = \
        argparse.ArgumentParser(
            prog=progname,
            description='Check PyTorch models for (overtly) unsafe code',
            epilog='Copyright (C) Joel Eriksson <je@clevcode.org> 2023')
    parser.add_argument(
        'target',
        metavar='PATH',
        help='Files and/or directories',
        type=str, nargs='+')
    parser.add_argument(
        '-t', '--trace',
        help='Safely trace the Pickle VM execution',
        action='store_true')
    parser.add_argument(
        '-d', '--dump',
        help='Dump the disassembly of all pickled bytecode',
        action='store_true')
    args = parser.parse_args()

    options = {}
    if args.trace:
        options['trace'] = True
    if args.dump:
        options['dump'] = True

    for path in args.target:
        # Check any files that were specified explicitly, since PyTorch
        # models may use any file extension even though .pt and .pth are
        # the ones that are most commonly used.
        if not os.path.exists(path):
            print(colorama.Fore.RED+f'{path} does not exist')
            success = False
        elif os.path.isfile(path):
            process(path)
        elif os.path.isdir(path):
            # For directories, just check .pt/.pth and .pkl/.pickle files
            for root, _, files in os.walk(path):
                for file in files:
                    ext = file.split('.')[-1].lower()
                    # NOTE! Other file extensions may be used as well!
                    if ext in ['ckpt', 'pt', 'pth', 'pkl', 'pickle']:
                        process(os.path.join(root, file), options)

if __name__ == '__main__':
    main()
    sys.exit(0 if success else 1)



from fickling.analysis import check_safety
from fickling import pickle, tracing

import pickletools
import zipfile
import sys
import io

if sys.version_info >= (3, 9):
    from ast import unparse
else:
    from astunparse import unparse

def check(file_or_buf, options={}):
    pickled = pickle.Pickled.load(file_or_buf)
    if 'trace' in options:
        trace = tracing.Trace(pickle.Interpreter(pickled))
        print(unparse(trace.run()))
    result = check_safety(pickled)
    if 'dump' in options:
        f = io.BytesIO()
        pickled.dump(f)
        f.seek(0)
        pickletools.dis(f)
    return result

def check_pytorch(filename, options={}):
    result = True
    with zipfile.ZipFile(filename, 'r') as zipped:
        for filename in zipped.namelist():
            if filename.endswith('.pkl') or filename.endswith('.debug_pkl'):
                print(f'Embedded pickle file {filename}')
                with zipped.open(filename, 'r') as file:
                    if not check(file.read(), options):
                        result = False
            elif filename.endswith('.py'):
                # Not sure if this is used or just included for reference?!
                print(f'Embedded Python file {filename}')
            else:
                # Just ignore any data files for now
                pass
    return result

def check_file(filename, options={}):
    try:
        ext = filename.split('.')[-1].lower()
        if ext in ['pkl', 'pickle']:
            # If it's a .pkl/.pickle file, check it directly
            return check(open(filename, 'rb'), options)
        else:
            # Otherwise assume it's a PyTorch model file
            return check_pytorch(filename, options)
    except Exception as e:
        print(f'Exception {e} while checking {filename}')
        return False
